"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.useMinilinks = exports.minilinks = exports.MinilinkCollection = exports.Minilinks = exports.MinilinksGeneratorOptionsDefault = void 0;
const remove_1 = __importDefault(require("lodash/remove"));
const isEqual_1 = __importDefault(require("lodash/isEqual"));
const events_1 = __importDefault(require("events"));
const react_1 = require("react");
exports.MinilinksGeneratorOptionsDefault = {
    id: 'id',
    type_id: 'type_id',
    type: 'type',
    typed: 'typed',
    from_id: 'from_id',
    from: 'from',
    out: 'out',
    to_id: 'to_id',
    to: 'to',
    in: 'in',
    inByType: 'inByType',
    outByType: 'outByType',
    equal: (ol, nl) => {
        return ol.type_id == nl.type_id && ol.from_id == nl.from_id && ol.to_id == nl.to_id && (0, isEqual_1.default)(ol.value, nl.value);
    },
};
function Minilinks(options) {
    return function minilinks(linksArray = [], memory = {}) {
        if (!memory.options)
            memory.options = options;
        const types = memory.types || {};
        const byId = memory.byId || {};
        const links = memory.links || [];
        const newLinks = [];
        for (let l = 0; l < linksArray.length; l++) {
            if (byId[linksArray[l][options.id]]) {
                if (options.handler)
                    options.handler(byId[linksArray[l][options.id]], memory);
            }
            else {
                const link = Object.assign(Object.assign({}, linksArray[l]), { [options.typed]: [], [options.in]: [], [options.out]: [], [options.inByType]: {}, [options.outByType]: {} });
                byId[link[options.id]] = link;
                types[link[options.type_id]] = types[link[options.type_id]] || [];
                types[link[options.type_id]].push(link);
                links.push(link);
                newLinks.push(link);
            }
        }
        for (let l = 0; l < newLinks.length; l++) {
            const link = newLinks[l];
            if (byId[link[options.type_id]]) {
                link[options.type] = byId[link[options.type_id]];
                byId[link[options.type_id]][options.typed].push(link);
            }
            if (byId[link[options.from_id]]) {
                link[options.from] = byId[link[options.from_id]];
                byId[link[options.from_id]][options.out].push(link);
                byId[link[options.from_id]][options.outByType][link[options.type_id]] = byId[link[options.from_id]][options.outByType][link[options.type_id]] || [];
                byId[link[options.from_id]][options.outByType][link[options.type_id]].push(link);
            }
            if (byId[link[options.to_id]]) {
                link[options.to] = byId[link[options.to_id]];
                byId[link[options.to_id]][options.in].push(link);
                byId[link[options.to_id]][options.inByType][link[options.type_id]] = byId[link[options.to_id]][options.inByType][link[options.type_id]] || [];
                byId[link[options.to_id]][options.inByType][link[options.type_id]].push(link);
            }
            if (options.handler)
                options.handler(link, memory);
        }
        if (!memory.types)
            memory.types = types;
        if (!memory.byId)
            memory.byId = byId;
        if (!memory.links)
            memory.links = links;
        return memory;
    };
}
exports.Minilinks = Minilinks;
class MinilinkCollection {
    constructor(options, memory) {
        this.types = {};
        this.byId = {};
        this.byFrom = {};
        this.byTo = {};
        this.byType = {};
        this.links = [];
        this._updating = false;
        const _options = options || exports.MinilinksGeneratorOptionsDefault;
        this.types = this.byType = (memory === null || memory === void 0 ? void 0 : memory.types) || {};
        this.byId = (memory === null || memory === void 0 ? void 0 : memory.byId) || {};
        this.byFrom = (memory === null || memory === void 0 ? void 0 : memory.byFrom) || {};
        this.byTo = (memory === null || memory === void 0 ? void 0 : memory.byTo) || {};
        this.links = (memory === null || memory === void 0 ? void 0 : memory.links) || [];
        this.options = _options;
        this.emitter = new events_1.default();
    }
    add(linksArray) {
        var _a, _b, _c, _d, _e, _f;
        const { byId, byFrom, byTo, byType, links, options } = this;
        const anomalies = [];
        const errors = [];
        const newLinks = [];
        for (let l = 0; l < linksArray.length; l++) {
            if (!!byId[linksArray[l][options.id]])
                errors.push(new Error(`${linksArray[l][options.id]} can't add because already exists in collection`));
            if (byId[linksArray[l][options.id]]) {
                if (options.handler)
                    options.handler(byId[linksArray[l][options.id]], this);
            }
            else {
                const link = Object.assign(Object.assign({}, linksArray[l]), { [options.typed]: [], [options.in]: [], [options.out]: [], [options.inByType]: {}, [options.outByType]: {} });
                byId[link[options.id]] = link;
                if (link[options.from_id]) {
                    if (byFrom[link[options.from_id]])
                        byFrom[link[options.from_id]].push(link);
                    else
                        byFrom[link[options.from_id]] = [link];
                }
                if (link[options.to_id]) {
                    if (byTo[link[options.to_id]])
                        byTo[link[options.to_id]].push(link);
                    else
                        byTo[link[options.to_id]] = [link];
                }
                if (link[options.type_id]) {
                    if (byType[link[options.type_id]])
                        byType[link[options.type_id]].push(link);
                    else
                        byType[link[options.type_id]] = [link];
                }
                if ((_a = byType[link[options.id]]) === null || _a === void 0 ? void 0 : _a.length) {
                    for (let i = 0; i < ((_b = byType[link[options.id]]) === null || _b === void 0 ? void 0 : _b.length); i++) {
                        const dep = byType[link[options.id]][i];
                        dep.type = link;
                        link[options.typed].push(dep);
                    }
                }
                if ((_c = byFrom[link[options.id]]) === null || _c === void 0 ? void 0 : _c.length) {
                    for (let i = 0; i < ((_d = byFrom[link[options.id]]) === null || _d === void 0 ? void 0 : _d.length); i++) {
                        const dep = byFrom[link[options.id]][i];
                        dep.from = link;
                        link[options.out].push(dep);
                    }
                }
                if ((_e = byTo[link[options.id]]) === null || _e === void 0 ? void 0 : _e.length) {
                    for (let i = 0; i < ((_f = byTo[link[options.id]]) === null || _f === void 0 ? void 0 : _f.length); i++) {
                        const dep = byTo[link[options.id]][i];
                        dep.to = link;
                        link[options.in].push(dep);
                    }
                }
                links.push(link);
                newLinks.push(link);
            }
        }
        for (let l = 0; l < newLinks.length; l++) {
            const link = newLinks[l];
            if (byId[link[options.type_id]]) {
                link[options.type] = byId[link[options.type_id]];
                byId[link[options.type_id]][options.typed].push(link);
            }
            else if (link[options.type_id])
                anomalies.push(new Error(`${link[options.id]} link.type_id ${link[options.type_id]} not founded`));
            if (byId[link[options.from_id]]) {
                link[options.from] = byId[link[options.from_id]];
                byId[link[options.from_id]][options.out].push(link);
                byId[link[options.from_id]][options.outByType][link[options.type_id]] = byId[link[options.from_id]][options.outByType][link[options.type_id]] || [];
                byId[link[options.from_id]][options.outByType][link[options.type_id]].push(link);
            }
            else if (link[options.from_id])
                anomalies.push(new Error(`${link[options.id]} link.from_id ${link[options.from_id]} not founded`));
            if (byId[link[options.to_id]]) {
                link[options.to] = byId[link[options.to_id]];
                byId[link[options.to_id]][options.in].push(link);
                byId[link[options.to_id]][options.inByType][link[options.type_id]] = byId[link[options.to_id]][options.inByType][link[options.type_id]] || [];
                byId[link[options.to_id]][options.inByType][link[options.type_id]].push(link);
            }
            else if (link[options.to_id])
                anomalies.push(new Error(`${link[options.id]} link.to_id ${link[options.to_id]} not founded`));
            if (options.handler)
                options.handler(link, this);
            if (!this._updating)
                this.emitter.emit('added', link);
        }
        return {
            anomalies,
            errors,
        };
    }
    remove(idsArray) {
        var _a, _b, _c, _d, _e, _f;
        const { byId, byFrom, byTo, byType, types, links, options } = this;
        const anomalies = [];
        const errors = [];
        const newLinks = [];
        for (let l = 0; l < idsArray.length; l++) {
            const id = idsArray[l];
            if (!byId[id])
                errors.push(new Error(`${id} can't delete because not exists in collection`));
            for (let i = 0; i < ((_a = byFrom === null || byFrom === void 0 ? void 0 : byFrom[id]) === null || _a === void 0 ? void 0 : _a.length); i++) {
                const dep = (_b = byFrom === null || byFrom === void 0 ? void 0 : byFrom[id]) === null || _b === void 0 ? void 0 : _b[i];
                dep[options.from] = undefined;
            }
            byFrom === null || byFrom === void 0 ? true : delete byFrom[id];
            for (let i = 0; i < ((_c = byTo === null || byTo === void 0 ? void 0 : byTo[id]) === null || _c === void 0 ? void 0 : _c.length); i++) {
                const dep = (_d = byTo === null || byTo === void 0 ? void 0 : byTo[id]) === null || _d === void 0 ? void 0 : _d[i];
                dep[options.to] = undefined;
            }
            byTo === null || byTo === void 0 ? true : delete byTo[id];
            for (let i = 0; i < ((_e = byType === null || byType === void 0 ? void 0 : byType[id]) === null || _e === void 0 ? void 0 : _e.length); i++) {
                const dep = (_f = byType === null || byType === void 0 ? void 0 : byType[id]) === null || _f === void 0 ? void 0 : _f[i];
                dep[options.type] = undefined;
            }
            byType === null || byType === void 0 ? true : delete byType[id];
            byId === null || byId === void 0 ? true : delete byId[id];
            if (!this._updating)
                this.emitter.emit('removed', id);
        }
        (0, remove_1.default)(links, l => idsArray.includes(l[options.id]));
        return {
            anomalies,
            errors,
        };
    }
    apply(linksArray) {
        const { byId, byFrom, byTo, byType, types, links, options } = this;
        const toAdd = [];
        const toUpdate = [];
        const beforeUpdate = {};
        const toRemove = [];
        const _byId = {};
        for (let l = 0; l < linksArray.length; l++) {
            const link = linksArray[l];
            const old = byId[link.id];
            if (!old)
                toAdd.push(link);
            else if (!options.equal(old, link)) {
                toUpdate.push(link);
                beforeUpdate[link.id] = old;
            }
            _byId[link.id] = link;
        }
        for (let l = 0; l < links.length; l++) {
            const link = links[l];
            if (!_byId[link.id])
                toRemove.push(link);
        }
        const r1 = this.remove(toRemove.map(l => l[options.id]));
        const a1 = this.add(toAdd);
        this._updating = true;
        const r2 = this.remove(toUpdate.map(l => l[options.id]));
        const a2 = this.add(toUpdate);
        for (let i = 0; i < toUpdate.length; i++) {
            const l = toUpdate[i];
            this.emitter.emit('updated', beforeUpdate[l.id], byId[l.id]);
        }
        this._updating = false;
        return { errors: [...r1.errors, ...a1.errors, ...r2.errors, ...a2.errors], anomalies: [...r1.anomalies, ...a1.anomalies, ...r2.anomalies, ...a2.anomalies] };
    }
}
exports.MinilinkCollection = MinilinkCollection;
exports.minilinks = Minilinks(exports.MinilinksGeneratorOptionsDefault);
function useMinilinks() {
    const mlRef = (0, react_1.useRef)((0, react_1.useMemo)(() => {
        return new MinilinkCollection();
    }, []));
    const ml = mlRef.current;
    return { ml, ref: mlRef };
}
exports.useMinilinks = useMinilinks;
//# sourceMappingURL=minilinks.js.map