"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.handleOperation = exports.handlerOperations = exports.useRunner = exports.makePromiseResult = exports.api = void 0;
const debug_1 = __importDefault(require("debug"));
const client_1 = require("@deep-foundation/hasura/client");
const api_1 = require("@deep-foundation/hasura/api");
const apollo_boost_1 = require("apollo-boost");
const promise_1 = require("../promise");
const client_2 = require("../client");
const global_ids_1 = require("../global-ids");
const child_process_1 = require("child_process");
const axios_1 = __importDefault(require("axios"));
const crypto_1 = __importDefault(require("crypto"));
const SCHEMA = 'public';
const debug = (0, debug_1.default)('deepcase:eh');
exports.api = new api_1.HasuraApi({
    path: process.env.DEEPLINKS_HASURA_PATH,
    ssl: !!+process.env.DEEPLINKS_HASURA_SSL,
    secret: process.env.DEEPLINKS_HASURA_SECRET,
});
const client = (0, client_1.generateApolloClient)({
    path: `${process.env.DEEPLINKS_HASURA_PATH}/v1/graphql`,
    ssl: !!+process.env.DEEPLINKS_HASURA_SSL,
    secret: process.env.DEEPLINKS_HASURA_SECRET,
});
const deep = new client_2.DeepClient({
    apolloClient: client,
});
const portHash = {};
const containerHash = {};
function makePromiseResult(promise, resolvedTypeId, promiseResultTypeId, result, promiseReasonTypeId, handleInsertId) {
    return {
        from: {
            data: {
                from_id: promise.id,
                type_id: resolvedTypeId,
                to: {
                    data: {
                        type_id: promiseResultTypeId,
                        object: { data: { value: result } },
                    }
                }
            }
        },
        type_id: promiseReasonTypeId,
        to_id: handleInsertId
    };
}
exports.makePromiseResult = makePromiseResult;
;
const useRunner = ({ code, beforeLink, afterLink }) => __awaiter(void 0, void 0, void 0, function* () {
    var _a, _b;
    console.log("handler4: ");
    const runnerPort = 3020;
    const runnerImageAndTag = 'konard/deep-runner-js:main';
    const jwt = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJodHRwczovL2hhc3VyYS5pby9qd3QvY2xhaW1zIjp7IngtaGFzdXJhLWFsbG93ZWQtcm9sZXMiOlsibGluayJdLCJ4LWhhc3VyYS1kZWZhdWx0LXJvbGUiOiJsaW5rIiwieC1oYXN1cmEtdXNlci1pZCI6IjI0In0sImlhdCI6MTY0MDM5MDY1N30.l8BHkbl0ne3yshcF73rgPVR-Sskr0hHECr_ZsJyCdxA';
    const data = jwt + code;
    const containerName = crypto_1.default.createHash('md5').update(data).digest("hex");
    let port;
    if (!portHash[containerName]) {
        port = Math.floor(Math.random() * 1000) + 1000;
        portHash[containerName] = port;
        containerHash[port] = containerName;
        const startDocker = `docker run --name ${containerName} -p ${port}:${runnerPort} -d ${runnerImageAndTag} && npx -q wait-on --timeout 20000 http://localhost:${port}/healthz`;
        let startResult;
        try {
            startResult = yield (0, child_process_1.execSync)(startDocker).toString();
        }
        catch (e) {
            startResult = e.stderr;
        }
        console.log('startResult', startResult.toString());
        console.log(1);
        while (startResult.indexOf('port is already arllocated') !== -1) {
            console.log(2);
            containerHash[port] = 'broken';
            port = Math.floor(Math.random() * 1000) + 1000;
            containerHash[port] = containerName;
            const REstartDocker = `docker run --name ${containerName} -p ${port}:3020 -d ${runnerImageAndTag} && npx -q wait-on --timeout 20000 http://localhost:${port}/healthz`;
            const startResult = yield (0, child_process_1.execSync)(REstartDocker).toString();
            console.log('REstartResult', startResult);
        }
        console.log(3);
        if (startResult.indexOf('is already in use by container') !== -1) {
            console.log(4);
            const REstartDocker = `docker stop ${containerName} && docker rm ${containerName} && docker run --name ${containerName} -p ${port}:3020 -d ${runnerImageAndTag} && npx -q wait-on --timeout 10000 http://localhost:${port}/healthz`;
            try {
                yield (0, child_process_1.execSync)(REstartDocker).toString();
            }
            catch (e) {
                console.log(e === null || e === void 0 ? void 0 : e.status);
            }
        }
        console.log(5);
        const initResult = yield axios_1.default.post(`http://localhost:${port}/init`, { params: { code, jwt } });
        console.log('initResult', initResult.data);
        console.log(6);
    }
    else {
        port = portHash[containerName];
    }
    console.log(7);
    try {
        const result = yield axios_1.default.post(`http://localhost:${port}/call`, { params: { beforeLink, afterLink } });
        if ((_a = result === null || result === void 0 ? void 0 : result.data) === null || _a === void 0 ? void 0 : _a.resolved) {
            return result.data.resolved;
        }
        return Promise.reject((_b = result === null || result === void 0 ? void 0 : result.data) === null || _b === void 0 ? void 0 : _b.rejected);
    }
    catch (e) {
        console.log('e', e);
    }
});
exports.useRunner = useRunner;
exports.handlerOperations = {
    Insert: 'HandleInsert',
    Update: 'HandleUpdate',
    Delete: 'HandleDelete',
};
function handleOperation(operation, oldLink, newLink) {
    var _a, _b, _c, _d, _e;
    return __awaiter(this, void 0, void 0, function* () {
        const current = newLink !== null && newLink !== void 0 ? newLink : oldLink;
        const currentLinkId = current.id;
        const currentTypeId = current.type_id;
        const handlerTypeId = yield deep.id('@deep-foundation/core', 'Handler');
        const handleOperationTypeId = yield deep.id('@deep-foundation/core', exports.handlerOperations[operation]);
        const queryString = `query SELECT_CODE($typeId: bigint) { links(where: {
          type_id: { _eq: ${yield deep.id('@deep-foundation/core', 'SyncTextFile')} },
          in: {
            from_id: { _eq: ${yield deep.id('@deep-foundation/core', 'JSExecutionProvider')} },
            type_id: { _eq: ${handlerTypeId} },
            in: {
              _or: [
                {
                  from: {
                    type_id: { _eq: $typeId },
                  }
                },
              ],
              type_id: { _eq: ${handleOperationTypeId} },
            }
          }
        }) {
          id
          value
          in(where: { type_id: { _eq: ${handlerTypeId} } }) {
            id
            in(where: { type_id: { _eq: ${handleOperationTypeId} } }) {
              id
            }
          }
        } }`;
        const query = (0, apollo_boost_1.gql) `${queryString}`;
        const variables = {
            typeId: currentTypeId
        };
        const handlersResult = yield client.query({ query, variables });
        const promises = [];
        const handleInsertsIds = [];
        const handlersWithCode = (_a = handlersResult === null || handlersResult === void 0 ? void 0 : handlersResult.data) === null || _a === void 0 ? void 0 : _a.links;
        if ((handlersWithCode === null || handlersWithCode === void 0 ? void 0 : handlersWithCode.length) > 0) {
            console.log("handlersWithCode: ", JSON.stringify(handlersWithCode, null, 2));
            console.log(handlersWithCode === null || handlersWithCode === void 0 ? void 0 : handlersWithCode.length);
            for (const handlerWithCode of handlersWithCode) {
                const code = (_b = handlerWithCode === null || handlerWithCode === void 0 ? void 0 : handlerWithCode.value) === null || _b === void 0 ? void 0 : _b.value;
                const handleInsertId = (_e = (_d = (_c = handlerWithCode === null || handlerWithCode === void 0 ? void 0 : handlerWithCode.in) === null || _c === void 0 ? void 0 : _c[0]) === null || _d === void 0 ? void 0 : _d.in) === null || _e === void 0 ? void 0 : _e[0].id;
                if (code) {
                    try {
                        promises.push(() => (0, exports.useRunner)({ code, beforeLink: oldLink, afterLink: newLink }));
                        handleInsertsIds.push(handleInsertId);
                    }
                    catch (error) {
                        debug('error', error);
                    }
                }
            }
            const resolvedTypeId = yield deep.id('@deep-foundation/core', 'Resolved');
            const rejectedTypeId = yield deep.id('@deep-foundation/core', 'Rejected');
            const promiseResultTypeId = yield deep.id('@deep-foundation/core', 'PromiseResult');
            const promiseReasonTypeId = yield deep.id('@deep-foundation/core', 'PromiseReason');
            const promise = yield (0, promise_1.findPromiseLink)({
                id: currentLinkId, client: deep.apolloClient,
                Then: yield deep.id('@deep-foundation/core', 'Then'),
                Promise: yield deep.id('@deep-foundation/core', 'Promise'),
                Resolved: resolvedTypeId,
                Rejected: rejectedTypeId,
                Results: false,
            });
            console.log('promise: ', promise);
            if (promise) {
                console.log("promises.length: ", promises.length);
                yield Promise.allSettled(promises.map((p) => p()))
                    .then((values) => __awaiter(this, void 0, void 0, function* () {
                    var _f;
                    console.log("values: ", values);
                    const promiseResults = [];
                    for (let i = 0; i < values.length; i++) {
                        const value = values[i];
                        const handleInsertId = handleInsertsIds[i];
                        if (value.status == 'fulfilled') {
                            const result = value.value;
                            console.log("result: ", result);
                            const promiseResult = makePromiseResult(promise, resolvedTypeId, promiseResultTypeId, result, promiseReasonTypeId, handleInsertId);
                            promiseResults.push(promiseResult);
                        }
                        if (value.status == 'rejected') {
                            const error = value.reason;
                            console.log("error: ", error);
                            const promiseResult = makePromiseResult(promise, rejectedTypeId, promiseResultTypeId, error, promiseReasonTypeId, handleInsertId);
                            promiseResults.push(promiseResult);
                        }
                    }
                    try {
                        yield deep.insert(promiseResults, { name: 'IMPORT_PROMISES_RESULTS' });
                        console.log("inserted promiseResults: ", JSON.stringify(promiseResults, null, 2));
                    }
                    catch (e) {
                        console.log('promiseResults insert error: ', (_f = e === null || e === void 0 ? void 0 : e.message) !== null && _f !== void 0 ? _f : e);
                    }
                }));
            }
        }
    });
}
exports.handleOperation = handleOperation;
exports.default = (req, res) => __awaiter(void 0, void 0, void 0, function* () {
    var _c, _d, _e, _f, _g, _h, _j;
    try {
        const event = (_c = req === null || req === void 0 ? void 0 : req.body) === null || _c === void 0 ? void 0 : _c.event;
        const operation = event === null || event === void 0 ? void 0 : event.op;
        if (operation === 'INSERT' || operation === 'UPDATE' || operation === 'DELETE') {
            const oldRow = (_d = event === null || event === void 0 ? void 0 : event.data) === null || _d === void 0 ? void 0 : _d.old;
            const newRow = (_e = event === null || event === void 0 ? void 0 : event.data) === null || _e === void 0 ? void 0 : _e.new;
            if (oldRow) {
                const queryResult = yield deep.select({
                    id: { _eq: oldRow.id },
                }, {
                    returning: `value`,
                });
                oldRow.value = (_g = (_f = queryResult.data) === null || _f === void 0 ? void 0 : _f[0]) === null || _g === void 0 ? void 0 : _g.value;
            }
            if (newRow) {
                const queryResult = yield deep.select({
                    id: { _eq: newRow.id },
                }, {
                    returning: `value`,
                });
                newRow.value = (_j = (_h = queryResult.data) === null || _h === void 0 ? void 0 : _h[0]) === null || _j === void 0 ? void 0 : _j.value;
            }
            debug('event', JSON.stringify(event, null, 2));
            debug('oldRow', oldRow);
            debug('newRow', newRow);
            const current = operation === 'DELETE' ? oldRow : newRow;
            const typeId = current.type_id;
            try {
                if (operation === 'INSERT') {
                    yield handleOperation('Insert', oldRow, newRow);
                }
                else if (operation === 'UPDATE') {
                }
                else if (operation === 'DELETE') {
                    yield handleOperation('Delete', oldRow, newRow);
                }
                if (operation === 'INSERT' && !global_ids_1.DENIED_IDS.includes(current.type_id) && global_ids_1.ALLOWED_IDS.includes(current.type_id)) {
                    debug('resolve', current.id);
                    yield (0, promise_1.resolve)({
                        id: current.id, client,
                        Then: yield deep.id('@deep-foundation/core', 'Then'),
                        Promise: yield deep.id('@deep-foundation/core', 'Promise'),
                        Resolved: yield deep.id('@deep-foundation/core', 'Resolved'),
                        Rejected: yield deep.id('@deep-foundation/core', 'Rejected'),
                        Results: false,
                    });
                }
                return res.status(200).json({});
            }
            catch (error) {
                debug('error', error);
                if (operation === 'INSERT' && !global_ids_1.DENIED_IDS.includes(current.type_id) && global_ids_1.ALLOWED_IDS.includes(current.type_id)) {
                    debug('reject', current.id);
                    yield (0, promise_1.reject)({
                        id: current.id, client,
                        Then: yield deep.id('@deep-foundation/core', 'Then'),
                        Promise: yield deep.id('@deep-foundation/core', 'Promise'),
                        Resolved: yield deep.id('@deep-foundation/core', 'Resolved'),
                        Rejected: yield deep.id('@deep-foundation/core', 'Rejected'),
                        Results: false,
                    });
                }
            }
            return res.status(500).json({ error: 'notexplained' });
        }
        return res.status(500).json({ error: 'operation can be only INSERT or UPDATE' });
    }
    catch (error) {
        return res.status(500).json({ error: error.toString() });
    }
});
//# sourceMappingURL=links.js.map